from __future__ import annotations

import json
from collections.abc import Callable, Mapping, Sequence
from typing import Any

from .protocols import (
    CoreNamespace,
    CrossLayerProps,
    ErrorDetails,
    ErrorObject,
    LayerDescription,
    LogId,
    LogLevel,
    LogLevelNames,
)


def get_log_level_name(log_level: LogLevel) -> str:
    if log_level == LogLevel.TRACE:
        return "TRACE"
    if log_level == LogLevel.DEBUG:
        return "DEBUG"
    if log_level == LogLevel.INFO:
        return "INFO"
    if log_level == LogLevel.WARN:
        return "WARN"
    if log_level == LogLevel.ERROR:
        return "ERROR"
    if log_level == LogLevel.SILENT:
        return "SILENT"
    raise ValueError(f"Unhandled log level {log_level}")


def get_log_level_number(log_level: LogLevelNames) -> int:
    if log_level == LogLevelNames.trace:
        return LogLevel.TRACE.value
    if log_level == LogLevelNames.debug:
        return LogLevel.DEBUG.value
    if log_level == LogLevelNames.info:
        return LogLevel.INFO.value
    if log_level == LogLevelNames.warn:
        return LogLevel.WARN.value
    if log_level == LogLevelNames.error:
        return LogLevel.ERROR.value
    if log_level == LogLevelNames.silent:
        return LogLevel.SILENT.value
    raise ValueError(f"Unhandled log level {log_level}")


def _get_layer_key(layer: LayerDescription) -> str:
    if isinstance(layer, list):
        return "-".join(layer)
    return str(layer)


def get_layers_unavailable(
    all_layers: Sequence[LayerDescription],
) -> Callable[[str], list[str]]:
    layer_to_choices: dict[str, list[str]] = {}
    for idx, layer in enumerate(all_layers):
        anti_layers = list(all_layers[idx + 1 :])
        if isinstance(layer, list):
            for i, composite_layer in enumerate(layer):
                nested_anti = layer[i + 1 :]
                layer_to_choices[composite_layer] = [
                    choice for choice in _flatten_layers(anti_layers + nested_anti)
                ]
        else:
            layer_to_choices[_get_layer_key(layer)] = [
                choice for choice in _flatten_layers(anti_layers)
            ]

    def resolver(layer_name: str) -> list[str]:
        if layer_name not in layer_to_choices:
            raise ValueError(f"{layer_name} is not a valid layer choice")
        return layer_to_choices[layer_name]

    return resolver


def _flatten_layers(layers: Sequence[LayerDescription]) -> list[str]:
    result: list[str] = []
    for layer in layers:
        if isinstance(layer, list):
            result.extend(layer)
        else:
            result.append(layer)
    return result


def is_config(obj: Any) -> bool:
    if isinstance(obj, str):
        return False
    try:
        core = obj[CoreNamespace.root]  # type: ignore[index]
        _ = core["layer_order"]
        return True
    except Exception:
        return False


def validate_config(config: Mapping[str, Any]) -> None:
    def _require(path: list[str | CoreNamespace], type_: type | None = None) -> None:
        cur: Any = config
        for key in path:
            key_s = key.value if isinstance(key, CoreNamespace) else key
            if not isinstance(cur, Mapping) or key_s not in cur:
                raise ValueError(f"{'.'.join(map(str, path))} was not found in config")
            cur = cur[key_s]
        if type_ is not None and not isinstance(cur, type_):
            raise ValueError(
                f"{'.'.join(map(str, path))} must be of type {type_.__name__}"
            )

    _require(["environment"])
    _require(["system_name"])
    _require([CoreNamespace.root, "apps"])
    if not isinstance(config[CoreNamespace.root.value]["apps"], list):
        raise ValueError(f"{CoreNamespace.root.value}.apps must be an array")
    _require([CoreNamespace.root, "layer_order"])
    if not isinstance(config[CoreNamespace.root.value]["layer_order"], list):
        raise ValueError(f"{CoreNamespace.root.value}.layer_order must be an array")
    _require([CoreNamespace.root, "logging", "log_level"])
    _require([CoreNamespace.root, "logging", "log_format"])
    for app in config[CoreNamespace.root.value]["apps"]:
        if not isinstance(app, Mapping) or "name" not in app:
            raise ValueError("A configured app does not have a name.")


def combine_cross_layer_props(
    a: CrossLayerProps, b: CrossLayerProps
) -> CrossLayerProps:
    a_ids = list(a.get("logging", {}).get("ids", [])) if a.get("logging") else []
    b_ids = list(b.get("logging", {}).get("ids", [])) if b.get("logging") else []

    existing = {f"{k}:{v}": True for obj in a_ids for k, v in obj.items()}
    unique: list[LogId] = []
    for obj in b_ids:
        for k, v in obj.items():
            key = f"{k}:{v}"
            if key not in existing:
                unique.append({k: v})
    final_ids = a_ids + unique
    logging_other = dict(a.get("logging", {}))
    logging_other.pop("ids", None)
    result: CrossLayerProps = {"logging": {"ids": final_ids, **logging_other}}
    return result


def _convert_error_to_cause(error: Exception, code: str, message: str) -> ErrorObject:
    err: ErrorDetails = {"code": code, "message": message or str(error)}
    if getattr(error, "message", None):
        err["details"] = str(error)
    cause = getattr(error, "__cause__", None)
    if isinstance(cause, Exception):
        cause_obj = _convert_error_to_cause(cause, "NestedError", str(cause))
        err["cause"] = cause_obj["error"]
    return {"error": err}


def create_error_object(
    code: str, message: str, error: Any | None = None
) -> ErrorObject:
    base: ErrorObject = {"error": {"code": code, "message": message}}
    if error is None:
        return base
    if isinstance(error, Exception):
        details: ErrorObject = {
            "error": {
                "details": str(error),
                "message": message,
                "code": code,
            }
        }
        cause = getattr(error, "__cause__", None)
        if isinstance(cause, Exception):
            cause_obj = _convert_error_to_cause(cause, "CauseError", str(cause))
            details["error"]["cause"] = cause_obj["error"]
        return _merge(base, details)
    if isinstance(error, str):
        return _merge(base, {"error": {"details": error}})
    if isinstance(error, Mapping):
        try:
            json.dumps(error)
            return _merge(base, {"error": {"data": dict(error)}})
        except Exception:
            return _merge(base, {"error": {"details": str(error)}})
    return _merge(base, {"error": {"details": str(error)}})


def is_error_object(value: Any) -> bool:
    return isinstance(value, Mapping) and "error" in value


def _merge(a: Mapping[str, Any], b: Mapping[str, Any]) -> dict[str, Any]:
    out: dict[str, Any] = dict(a)
    for k, v in b.items():
        if k in out and isinstance(out[k], Mapping) and isinstance(v, Mapping):
            out[k] = _merge(out[k], v)  # type: ignore[assignment]
        else:
            out[k] = v  # type: ignore[assignment]
    return out


def get_namespace(package_name: str, app: str | None = None) -> str:
    if app:
        return f"{package_name}/{app}"
    return package_name


def do_nothing_fetcher(model: Any, primary_key: Any) -> Any:  # noqa: ARG001
    return primary_key



from .entries import SystemProps, load_system
from .globals.libs import (  # noqa: F401
    extract_cross_layer_props,
)
from .globals.logging import (
    composite_logger,
    console_log_full,
    console_log_json,
    console_log_simple,
    log_tcp,
    standard_logger,
)
from .libs import (  # noqa: F401
    combine_cross_layer_props,
    create_error_object,
    get_layers_unavailable,
    get_log_level_name,
    get_log_level_number,
    is_config,
    validate_config,
)
from .protocols import (  # noqa: F401
    App,
    AppLayer,
    CommonContext,
    Config,
    CoreConfig,
    CoreLoggingConfig,
    CoreNamespace,
    CrossLayerProps,
    FeaturesContext,
    FunctionLogger,
    HighLevelLogger,
    LayerContext,
    LayerDescription,
    LayerLogger,
    LogFormat,
    Logger,
    LogId,
    LogLevel,
    LogLevelNames,
    LogMessage,
    LogMethod,
    RootLogger,
)

__all__ = [
    "composite_logger",
    "console_log_full",
    "console_log_json",
    "console_log_simple",
    "load_system",
    "log_tcp",
    "standard_logger",
]



from __future__ import annotations

import os
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any

from box import Box

from .globals import features as globals_features
from .globals import name as globals_name
from .globals import services as globals_services
from .layers import features as layers_features
from .layers import name as layers_name
from .layers import services as layers_services


@dataclass(frozen=True)
class SystemProps:
    environment: str
    config: Mapping[str, Any] | None = None


async def load_system(props: SystemProps):
    global_services = globals_services.create(
        {
            "environment": props.environment,
            "working_directory": os.getcwd(),
        }
    )
    global_features = globals_features.create(
        Box(
            {
                "services": {
                    globals_name: global_services,
                }
            }
        )
    )
    globals_context = await global_features.load_globals(
        props.config or props.environment
    )

    # layers

    the_layers_services = layers_services.create()
    the_layers_features = layers_features.create(
        Box(
            {
                **globals_context,
                "services": {
                    layers_name: the_layers_services,
                },
            }
        )
    )
    layers_loaded = await the_layers_features.load_layers()
    try:
        if "services" in layers_loaded and layers_name in layers_loaded.services:
            del layers_loaded.services[layers_name]
    except Exception:  # noqa: S110
        pass
    return layers_loaded



from __future__ import annotations

from typing import Any

from ..libs import is_config, validate_config
from ..protocols import CommonContext, CoreNamespace, FeaturesContext

globals_name = CoreNamespace.globals.value


class GlobalsFeatures:
    def __init__(self, context: FeaturesContext):
        self.context = context

    async def load_globals(self, environment_or_config: Any):
        services = self.context.services[globals_name]
        if not services:
            raise RuntimeError(f"Services for {globals_name} not found")
        config = (
            environment_or_config
            if is_config(environment_or_config)
            else services.load_config()
        )
        validate_config(config)
        common_globals: CommonContext = {
            "config": config,
            "root_logger": services.get_root_logger(),
            "constants": services.get_constants(),
        }
        return common_globals


def create(context: FeaturesContext) -> GlobalsFeatures:
    return GlobalsFeatures(context)



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from ..protocols import ServicesContext


class LayersServices:
    def get_model_props(self, context: ServicesContext):
        raise NotImplementedError("Model support not implemented in Python port")

    def load_layer(
        self, app: Mapping[str, Any], layer: str, context: Mapping[str, Any]
    ):
        constructor = app.get(layer)
        if not constructor or "create" not in constructor:
            return None
        instance = constructor.create(context)
        if instance is None:
            raise RuntimeError(
                f"App {app.get('name')} did not return an instance layer {layer}"
            )
        return instance


def create() -> LayersServices:
    return LayersServices()



from __future__ import annotations

from collections.abc import Mapping
from copy import deepcopy
from typing import Any

from box import Box

from ..globals.libs import extract_cross_layer_props
from ..libs import get_layers_unavailable
from ..protocols import (
    CommonContext,
    CoreNamespace,
    FeaturesContext,
)


class LayersFeatures:
    def __init__(self, context: FeaturesContext):
        self.context = context

    def _get_layer_context(
        self, common_context: Mapping[str, Any], layer: Mapping[str, Any] | None
    ) -> Mapping[str, Any]:
        if layer:
            merged = deepcopy(common_context)
            for k, v in layer.items():
                merged[k] = v
            return merged
        return common_context

    def _make_wrapped(self, f, logger_ids):
        def _inner2(*args, **kwargs):  # noqa: ARG001
            args_no_cross, cross = extract_cross_layer_props(list(args))
            return f(
                *args_no_cross,
                cross or {"logging": {"ids": logger_ids}},
            )

        return _inner2

    def _wrap_layer_functions(
        self,
        loaded_layer: Any,
        layer_logger,
        app_name: str,
        layer: str,
        ignore_layer_functions: Mapping[str, Any],
    ):
        def _iter_properties(obj: Any):
            if isinstance(obj, Mapping):
                for k, v in obj.items():
                    yield k, v
                return
            # Fallback to attribute-based discovery on class instances
            for name in dir(obj):
                if name.startswith("_"):
                    continue
                try:
                    attr = getattr(obj, name)
                except Exception:
                    continue
                yield name, attr

        out: dict[str, Any] = {}
        for property_name, func in _iter_properties(loaded_layer):
            if not callable(func):
                out[property_name] = func
                continue
            function_level_key = f"{app_name}.{layer}.{property_name}"
            if _get(ignore_layer_functions, function_level_key):
                out[property_name] = func
                continue

            def _make_inner(f):
                def _inner(log, *args, **kwargs):  # noqa: ARG001
                    return f(*args, **kwargs)

                return _inner

            wrapped = layer_logger._log_wrap(property_name, _make_inner(func))
            for attr in dir(func):
                try:  # noqa: SIM105
                    setattr(wrapped, attr, getattr(func, attr))
                except Exception:  # noqa: S110
                    pass
            out[property_name] = wrapped
        return out

    async def _load_composite_layer(
        self,
        app: Mapping[str, Any],
        composite_layers,
        common_context: Mapping[str, Any],
        previous_layer: Mapping[str, Any] | None,
        anti_layers_fn,
    ):
        result = {}
        for layer in composite_layers:
            layer_logger = (
                self.context.root_logger.get_logger(Box(common_context))
                .get_app_logger(app["name"])
                .get_layer_logger(layer)
            )
            the_context = dict(common_context)
            the_context["log"] = layer_logger
            wrapped_context = the_context
            loaded = self.context.services[CoreNamespace.layers.value].load_layer(
                app, layer, Box(wrapped_context)
            )
            if loaded:
                ignore_layer_functions = (
                    self.context.config[CoreNamespace.root.value].logging.get(
                        "ignore_layer_functions"
                    )
                    or {}
                )
                layer_level_key = f"{app['name']}.{layer}"
                should_ignore = _get(ignore_layer_functions, layer_level_key)
                final_layer = (
                    loaded
                    if should_ignore
                    else self._wrap_layer_functions(
                        loaded, layer_logger, app["name"], layer, ignore_layer_functions
                    )
                )
                result = {**result, layer: {app["name"]: final_layer}}
        return result

    async def _load_layer(
        self,
        app: Mapping[str, Any],
        current_layer: str,
        common_context: Mapping[str, Any],
        previous_layer: Mapping[str, Any] | None,
    ):
        layer_context1 = self._get_layer_context(common_context, previous_layer)
        layer_logger = (
            self.context.root_logger.get_logger(Box(layer_context1))
            .get_app_logger(app["name"])
            .get_layer_logger(current_layer)
        )
        layer_context = dict(layer_context1)
        layer_context["log"] = layer_logger

        logger_ids = layer_logger.get_ids()
        ignore_layer_functions = (
            self.context.config[CoreNamespace.root.value].logging.get(
                "ignore_layer_functions"
            )
            or {}
        )
        wrapped_context = {}
        for layer_key, layer_data in layer_context.items():
            if layer_key in (
                "_logging",
                "root_logger",
                "log",
                "constants",
                "config",
                "models",
                "get_models",
                "cruds",
            ):
                wrapped_context[layer_key] = layer_data
                continue
            if not isinstance(layer_data, Mapping):
                wrapped_context[layer_key] = layer_data
                continue
            final_layer_data = {}
            for domain_key, domain_value in layer_data.items():
                if not isinstance(domain_value, Mapping):
                    final_layer_data[domain_key] = domain_value
                    continue
                layer_level_key = f"{domain_key}.{layer_key}"
                if _get(ignore_layer_functions, layer_level_key):
                    final_layer_data[domain_key] = domain_value
                    continue
                domain_data = {}
                for property_name, func in domain_value.items():
                    if not callable(func):
                        domain_data[property_name] = func
                        continue
                    function_level_key = f"{domain_key}.{layer_key}.{property_name}"
                    if _get(ignore_layer_functions, function_level_key):
                        domain_data[property_name] = func
                        continue

                    wrapped_func = self._make_wrapped(func, logger_ids)
                    for attr in dir(func):
                        try:  # noqa: SIM105
                            setattr(wrapped_func, attr, getattr(func, attr))
                        except Exception:  # noqa: S110
                            pass
                    domain_data[property_name] = wrapped_func
                final_layer_data[domain_key] = domain_data
            wrapped_context[layer_key] = final_layer_data

        loaded = self.context.services[CoreNamespace.layers.value].load_layer(
            app, current_layer, Box(wrapped_context)
        )
        if not loaded:
            return {}
        layer_level_key = f"{app['name']}.{current_layer}"
        should_ignore = _get(ignore_layer_functions, layer_level_key)
        final_layer = (
            loaded
            if should_ignore
            else self._wrap_layer_functions(
                loaded, layer_logger, app["name"], current_layer, ignore_layer_functions
            )
        )
        return {current_layer: {app["name"]: final_layer}}

    async def load_layers(self):
        layers_in_order = self.context.config[CoreNamespace.root.value].layer_order
        anti_layers = get_layers_unavailable(layers_in_order)
        core_layers_to_ignore = [
            f"services.{CoreNamespace.layers.value}",
            f"services.{CoreNamespace.globals.value}",
            f"features.{CoreNamespace.layers.value}",
            f"features.{CoreNamespace.globals.value}",
        ]
        starting_context: CommonContext = {k: v for k, v in self.context.items() if k not in core_layers_to_ignore}  # type: ignore[return-value]
        apps = (
            self.context.config[CoreNamespace.root.value].get("apps")
            or self.context.config[CoreNamespace.root.value].get("domains")
            or []
        )
        existing_layers = starting_context
        for app in apps:
            previous_layer = {}
            for layer in layers_in_order:
                if isinstance(layer, list):
                    layer_instance = await self._load_composite_layer(
                        app,
                        layer,
                        {k: v for k, v in existing_layers.items() if k != "log"},
                        previous_layer,
                        anti_layers,
                    )
                else:
                    layer_instance = await self._load_layer(
                        app,
                        layer,
                        {k: v for k, v in existing_layers.items() if k != "log"},
                        previous_layer,
                    )
                if not layer_instance:
                    previous_layer = {}
                    continue
                new_context = {**existing_layers, **layer_instance}
                if "log" in new_context:
                    new_context = {k: v for k, v in new_context.items() if k != "log"}
                existing_layers = new_context
                previous_layer = layer_instance
        return Box(existing_layers)


def create(context: FeaturesContext) -> LayersFeatures:
    return LayersFeatures(context)


def _get(mapping: Mapping[str, Any], dotted: str, default: Any = None) -> Any:
    cur: Any = mapping
    for part in dotted.split("."):
        if isinstance(cur, Mapping) and part in cur:
            cur = cur[part]
        else:
            return default
    return cur
